#!/usr/bin/tclsh
#
# Velometrik Arbeitsstationsdeamon
#
# Ermöglicht die Steuerung einer Arbeitsstation mittels Browser.
#
# Author: Siegmar Müller
#
# Historie:
# ??.06.2019 Siegmar Müller Testversion fertiggestellt
# 05.05.2022 Siegmar Müller vmkstationd3 begonnen
# 01.11.2023 Siegmar Müller V 3.2 begonnen
#

set VERSION "3.2.0_U"

set ::dir [file dirname [info script]]
if {[string equal -length 1 $::dir "."]} {
    set dir "$env(PWD)[string range $::dir 1 end]"
}

# Start nur als root oder Angehöriger der Gruppe staff
set USER [exec id -n -u]
if {"$USER" != "root"} {
    set GROUPS [exec id -n -G]
    if {!("staff" in $GROUPS)} {
        puts stderr "vmkstationd must be executed by root or member of group staff"
        exit 1
    }
}

set CONF_PORT 80
set CONF_LOGFILE "$::dir/vmkstationd.log"; # Beim Start des Daemon über Kommndozeile änderbar.
set CONF_LOGLEVEL notice
if {[string index $VERSION end] in {A U}} {; # Alpha oder unfertige Version
    set CONF_LOGLEVEL debug
}
set ::PIEPSER 0; # Piepserplatine ist vorhanden (Wird nach erfolgreicher Initialisierung 1)
# Folgendes wird von der Initialisierung überschrieben:
set BGCOLOR #404040
set COLORCONTRAST 0
set JPEGOPTIONS [list -quality 60 -res 12 -grid no -frame 0]
set WS_URL "https://ws-1-1.velometrik.de/vmkservice"
set WS_VERSION ""; # wird beim Webservice abgefragt
set STATION_NR 0; # Nicht registrierte Station. Das wird bei der Registierung  überschrieben.
set using_defaults 1

# Workaround für ein Problem, das mit tcllib1.19 aufgetreten ist:
if {![info exists ::env(USER)]} {
    set ::env(USER) root
}
if {![info exists ::env(user)]} {
    set ::env(user) $::env(USER)
}
#Ende: Workaround


# Globale Daten
set ::SODIR "." ; # Verzeichnis der shared objects (wird noch ermittelt)
set ::BINDIR "/usr/local/bin" ; # Verzeichnis der ausführbaren Dateien (wird noch ermittelt)
set ::PORTALSERVER "" ; # Wird vom Webservice geliefert
# (siehe auch setup/vmkstationd.conf)
array set ::LEDFARBEN [list SCHWARZ 0 GRUEN 1 BLAU 2 HELLBLAU 3 ROT 4 GELBGRUEN 5 VIOLETT 6 WEISS 7]
# LAN beinhaltet WLAN
set ::STATUSLEVELS [list OFF POWER LAN INET WEBSERVICE ANGEMELDET MATTE]
array set ::STATUSFARBEN [list OFF SCHWARZ POWER ROT MATTE VIOLETT LAN BLAU INET HELLBLAU WEBSERVICE GRUEN ANGEMELDET WEISS]
# Der aktuelle Stationsstatus (OFF darf niemals 0 sein)
array set ::stationsstatus [list OFF 1 POWER 0 MATTE 0 LAN 0 INET 0 WEBSERVICE 0 ANGEMELDET 0]

#{{{ Konfiguration holen
#TODO Das sollte vorsichtshalber mit catch in einen eigenen namespace geholt werden,
#   um dann die entsprechenen Variablen zu übernehmen.
if {[file exists /usr/local/etc/vmkstationd.conf]} {
    source /usr/local/etc/vmkstationd.conf
    if {[info exists PORT]} {
        set CONF_PORT $PORT
    }
    if {[info exists LOGLEVEL]} {
        set CONF_LOGLEVEL $LOGLEVEL
    }
    set using_defaults 0
} elseif {[file exists $::dir/vmkstationd.conf]} {
    source $::dir/vmkstationd.conf
    if {[info exists PORT]} {
        set CONF_PORT $PORT
    }
    if {[info exists LOGLEVEL]} {
        set CONF_LOGLEVEL $LOGLEVEL
    }
    set using_defaults 0
}
#}}} Konfiguration holen

#{{{ Velobox Konfiguration
namespace eval VELOBOX_CONF {
    # Die erwarteten Variablen mit ihren Vorgaben
    set VERSION "0.0"
    set LANG "en"
    # vlbupdate
    set DOWNLOADSERVER ""
    set UPDATEDIR ""
    # List of additional hardware components
    # e.g.: HARDWARE=PIEPSER,JIT35
    set HARDWARE ""
    # Usage of this velobox
    # e.g. USAGE=
    set USAGE "bikefitting"
    # Implementation for USAGE
    # e.g. IMPLEMENTATION=velometrik
    set IMPLEMENTATION ""

    # Velobox Konfiguration einlesen
    # @param filename   Dateiname der Konfiguration
    proc readVeloboxConf {filename} {
        catch {
            set fd [open $filename "r"]
            while {[gets $fd line] >= 0} {
                if {[regexp {^[[:space:]]*([A-Za-z0-9]+)[[:space:]]*=[[:space:]]*([^[:space:]#]*)} $line match name value]} {
                    set [namespace current]::$name "$value"
                }
            }
            close $fd
        }
    }; # proc readVeloboxConf 

    # Tatsächliche Werte, sofern vorhanden, einlesen
    if {[file exists /usr/local/etc/velobox.conf]} {
        readVeloboxConf /usr/local/etc/velobox.conf
    }
}
#}}} Velobox Konfiguration

# Hilfe ausgeben und beenden
proc help {} {; #{{{
    puts ""
    puts "Velometrik Arbeitsstationsdeamon V $::VERSION"
    puts "Startparameter:"
    puts "    -p <port>     Start auf Port <port> (Vorgabe $::CONF_PORT)"
    puts "    -L <logfile>  Logfile (Vorgabe: $::CONF_LOGFILE)"
    puts "                  ('' => Ausgabe an stderr)"
    puts "    -l <loglevel> Loglevel (error warn notice info debug)"
    puts "                  (Vorgabe: $::CONF_LOGLEVEL)"
    puts "    -V        Version ausgeben und beenden"
    puts "    -h        Hilfe ausgeben und beenden"
    puts "    -?        Hilfe ausgeben und beenden"
    puts ""
    exit
    #}}}
}; # proc help


#{{{ Kommandozeilenparameter übernehmen
set status option
foreach arg $argv {
    switch -- $status {
        option {
            switch -- $arg {
                -p {
                    set status "-p"
                }
                -L {
                    set status "-L"
                }
                -l {
                    set status "-l"
                }
                -V {
                    puts $VERSION
                    exit
                }
                -h {
                    help
                    exit
                }
                -? {
                    help
                    exit
                }
                default {
                    puts stderr "Unbekannte Option: $arg"
                    exit 1
                }
            }; # switch $arg
        }
        "-p" {
            #TODO Integer?
            #TODO Rootrechte erforderlich?
            set CONF_PORT $arg
            set status option
        }
        "-L" {
            set CONF_LOGFILE $arg
            set status option
        }
        "-l" {
            if {[lsearch -exact  [list "error" "warn" "notice" "info" "debug"] $arg] < 0} {
                puts stderr "Unbekanntes Loglevel: '$arg'"
                exit 1
            }
            set CONF_LOGLEVEL $arg
            set status option
        }
    }; # switch $status
}
if {"$status" != "option"} {
    puts stderr "Wert für Option $status fehlt."
    exit 1
}
#}}} Kommandozeilenparameter sind übernommen


# Pakete laden
#TODO Die anderswo geladen Pakte auch hier laden oder dort, wo sie geladen werden
# Hinweis auf Debian-Package ergänzen
foreach {pkg_tcl pkg_debian} {http tcllib json tcllib Thread tcl-thread} {
    if {[catch {package require $pkg_tcl} errmsg]} {
        puts $errmsg
        puts "Is $pkg_debian installed?"
        exit 1
    }
}

# Letztes Logfile, falls vorhanden sichern.
proc logRestart {} {
    if {[file exists $::CONF_LOGFILE]} {
        set fd [open $::CONF_LOGFILE "a"]
        puts $fd "[clock format [clock seconds]] logRestart: ${::CONF_LOGFILE}~ wird erstellt"
        close $fd
        file copy -force $::CONF_LOGFILE ${::CONF_LOGFILE}~
        file delete $::CONF_LOGFILE
    }
    if {[llength [info procs srvLog]] > 0} {; # Kein Neustart, weil utils.tcl bereits geladen
        srvLog {} Notice "logRestart"
    }
    # Das nächste mal um Mitternacht
    after [expr {([clock add [clock scan 23:59] 1 minutes] - [clock seconds]) * 1000}] logRestart
}

logRestart


#{{{ Startkode holen und ausführen
# CONF_LOGFILE wird jetzt benötigt.
if {![file isdirectory [file dirname $::CONF_LOGFILE]]} {
    if {[catch {exec mkdir -p [file dirname $::CONF_LOGFILE]} errmsg]} {
        puts stderr "Logdir kann nicht angelegt werden: $errmsg"
        set ::CONF_LOGFILE ""
    }
}
if {[file isdirectory $::dir/start]} {
    foreach file [glob $::dir/start/*.tcl] {
        source $file
        #TODO Das geht schief, wenn utils.tcl nicht zuerst geholt wird.
        # => Dateien numerieren, wenn es mehrere werden.
        srvLog {} Notice "$file loaded"
    }
} else {
    puts stderr "Verzeichnis $::dir/start existiert nicht."
}
#}}} Startkode holen und ausführen

# Ab hier sind Logmeldungen möglich.
setLoglevel $CONF_LOGLEVEL

if {$using_defaults} {
    srvLog {} Notice "/usr/local/etc/vmkstationd.conf not found. Using defaults"
}

# Behandlung von Hintergrundfehlern
# Wird aufgerufen, wenn der Interpreter (bei vwait) in einen vorher nicht aufgefangenen Fehler läuft:
#TODO => Anlegen lohnt erst kurz vor vwait
# @param msg    Meldung vom TCL-Kern
proc bgerror {msg} {
    global errorInfo

    srvLog bgerror Error "$msg\n$errorInfo"
}; # proc bgerror


#{{{ TCL-Bibliotheken holen (binär und reines TCL)
set machine [exec uname -m]
if {[file isdirectory $::dir/bin/$machine]} {
    set ::BINDIR $::dir/bin/$machine
}
if {[file isdirectory $::dir/lib]} {
    if {[catch {
        if {[file isdirectory $::dir/lib/$machine]} {
            # shared Objects
            set ::SODIR $::dir/lib/$machine
            foreach file [glob -nocomplain $::SODIR/libtcl${tcl_version}*.so] {
                load $file
                srvLog {} Notice "$file loaded"
            }
        } else {
            srvLog {} Warn "$::dir/lib/${machine}/ doesn't exist."
        }
        #   reine TCL-Bibliotheken
        foreach file [glob -nocomplain $::dir/lib/*.tcl] {
            source $file
            srvLog {} Notice "$file loaded"
        }
    } msg]} {
        srvLog {} Error "Loading ${file}: $msg"
    }
} else {
   srvLog {} Warn "$dir/lib/ doesn't exist."
}
#}}} TCL-Bibliotheken holen


# Workaround zum Unterbinden der unnötig zeitverzögernden Hostabfrage bei nicht erreichbarem DNS-Server
proc ::websocket::fconfigure {args} {
    if {[lindex $args 1] in {-peername -sockname}} {
        upvar #0 clientcontext[lindex $args 0] context
        set ipaddr $context(ipaddr)
        set port $context(port)
        srvLog {Workaround} Debug "fconfigure $args => ${ipaddr}:$port"
        return [list $ipaddr $ipaddr $port]
    }
    return [::fconfigure {*}$args]
}; # proc ::websocket::fconfigure

# Fehlerbehandlung für Threads
proc threaderror {thread_id errorInfo} {
    srvLog [namespace current] Error "Thread id=$thread_id: $errorInfo"
}
thread::errorproc [namespace current]::threaderror 

##################################################################################################
proc tracecommand {command op} {
    srvLog trace Debug "$op: $command"
}
# trace add execution ::websocket::upgrade enterstep tracecommand
##################################################################################################

# Status setzen/zurücksetzen
# Schaltet den angegebenen Status ein oder aus
# und bringt anschließend den höchsten Statuslevel zur Anzeige,
# für den alle Vorgängerlevels eingeschaltet sind.
# @param statusname s. ::STATUSLEVELS
# @param on_off     0 oder 1
proc setStationsstatus {statusname {on_off 1}} {; #{{{
    set i_status [lsearch $::STATUSLEVELS $statusname]
    if {$i_status < 0} {; # interner Fehler
        return; # wird ignoriert.
    }
    set ::stationsstatus($statusname) $on_off
    set i_status 0
    foreach statuslevel $::STATUSLEVELS {
        if {!$::stationsstatus($statuslevel)} {
            break
        }
        incr i_status
    }
    incr i_status -1
    set status_neu [lindex $::STATUSLEVELS $i_status]
    if {$::PIEPSER} {
        ::piepser::setLED $::LEDFARBEN($::STATUSFARBEN($status_neu))
    }
    #TODO Statusänderung über Websocket verteilen => wsdomain einrichten
    srvLog {setStationsstatus} Notice "Stationsstatus: $status_neu"
    #}}}
}; # proc setStationsstatus 


# Piepser mit LEDs initialisieren
#TODO Das kommt in ein noch zu erstellendes Kernelmodul in lib/kernel
# Das scheitert, wenn es nicht auf einem Raspi ausgeführt wird.
# Es scheitert nicht, wenn es auf einem Raspi ohne Piepserplatine ausgeführt wird.
if {[catch {::piepser::gpioSetup} msg]} {
    srvLog gpioSetup Warn "Der Piepser wurde nicht aktiviert: $msg"
} else {
    set ::PIEPSER 1
    srvLog gpioSetup Notice "Piepser initialisiert"
    setStationsstatus POWER
}


#TODO Konfigurierbar:
set ::WSServer::HTTPD(default) "index.tcls"


# Kommunikation mit dem Webservice
# Prüft die Internetverbindung und meldet sich beim Webservice
namespace eval ::wscomm {; #{{{
    variable local_ip ""
    variable netzwerk_aktiv 0
    variable internet_verfuegbar 0
    variable webservice_verfuegbar 0
    variable after_id ""; # für die Wiederholungsversuche

    # Lokales Netzwerkinterface testen.
    # varmsg    Name der Variablen zum Hinterlegen der Meldung
    # => 1  Das Interface koennte funktionieren.
    #    0  Es wurde ein Fehler festgestellt.
    proc localIfcheck {msgvar} {; #{{{
        upvar $msgvar msg
        variable local_ip
    
        set driver ""
        if {[catch {
                foreach line [split [exec ip -4 address] "\n"] {
                    if {[regsub {^ *inet ([0-9.]*)/.*$} $line {\1} temp]} {; # temp hat local_ip
                        set local_ip $temp
                        if {"$driver" == "lo"} {
                            continue
                        } elseif {[string match "eth*" $driver]} {; # ... hat Vorrang
                            break
                        }
                    }
                    if {[regsub {^.* ([a-z][a-z0-9]*): .*$} $line {\1} temp]} {; # temp hat driver
                        set driver $temp
                        continue
                    }
                }
            } errmsg]} {
                set msg "LOCAL_IP kann nicht ermittelt werden: $errmsg"
                return 0
        } else {
            if {"$driver" == "lo"} {
                set msg "Keine Netzwerkverbindung"
                setStationsstatus LAN 0
                return 0
            } else {
                set msg "$driver: LOCAL_IP=$local_ip"
                setStationsstatus LAN
                return 1
            }
        }
        #}}}
    }; # proc localIfcheck


    # HTTP-Request ausfuehren
    # httpUrl   Vollstaendiger Request
    # datavar   Variable zum Hinterlegen der Antwort bzw. der Fehlermeldung
    # =>    1 Request wurde ausgefuehrt, datavar enthaelt das Ergebnis
    #       0 datavar enthaelt eine Fehlermeldung oder ist leer, wenn die Fehlermeldung zuvor schon uebergeben wurde
    proc execHttpRequest {httpUrl datavar} {; #{{{
        upvar $datavar data
        variable last_ws_error
        variable last_user_error

        set ok 0
        if {[catch {
            set tokenHttp [::http::geturl $httpUrl]; # Wirft evtl. eine Exception
            # Verbindung konnte hergestellt werden, falls hier angekommen.
            ::http::wait $tokenHttp
            set err_msg ""
            # http-Antwort auswerten
            switch [::http::status $tokenHttp] {
                ok {
                    if {[::http::ncode $tokenHttp] == 200} {
                        set data [::http::data $tokenHttp]
                        set ok 1
                    } else {
                        set err_msg "#WS: [::http::code $tokenHttp]"
                    }
                }
                eof {
                    set err_msg "#EOF: Connection closed by Webservice."
                }
                error {
                    set err_msg "#ERROR: [::http::error $tokenHttp]"
                }
            }
            ::http::cleanup $tokenHttp

            if {!$ok} {
                # Exception erst hier werfen, damit ::http::cleanup noch ausgefuehrt wird.
                error $err_msg
            }
        } catchResult]} {
            srvLog {} Error $catchResult
        }
        return $ok
        #}}}
    }; # execHttpRequest


    # Abarbeitung eines Webservice Requests
    # request   Bezeichnung des Requests
    # lquery    Parameter als Liste [list parameter wert ...]
    # datavar   Variable fuer das Ergebnis
    # =>    1 Request wurde ausgefuehrt, datavar enthaelt das Ergebnis
    #       0 datavar enthaelt eine Fehlermeldung oder ist leer, wenn die Fehlermeldung zuvor schon uebergeben wurde
    proc execWsRequest {request lquery datavar} {; #{{{
        upvar $datavar data

        set parameter "[::http::formatQuery {*}[join $lquery " "]]"
        if {[catch {
            srvLog {} Debug "Encrypting $parameter"
            set encrypted [vmkcrypt encrypt $parameter]
        } errmsg ]} {
            set data ""
            srvLog {} Error $errmsg
            return 0  
        }
        # Da keine Sonderzeichen mehr vorkommen koennen, kann das direkt zusammengebaut werden.
        #TODO Request kann nicht ausgeführt werden,
        # wenn STATION_NR == 0, d.h. nicht initialisiert ist.
        set parameter "sn=$::STATION_NR&req=$encrypted"
        srvLog {} Debug "execWsRequest: ${::WS_URL}/${request}?$parameter"
        return [execHttpRequest "${::WS_URL}/${request}?$parameter" data]
        #}}}
    }; # proc execWsRequest 


    # Ruft localIfcheck auf und prüft bei Erfolg den Internetzugang durch ping auf Google.
    # Wenn das erfolgreich war, wird dem Webserver die ermittelte local_ip mitgeteilt
    # und die zurückgegebene URL in ::PORTALSERVER gespeichert.
    # Falls diese Kette nicht fehlerfrei durchlaufen wurde, enthält ::PORTALSERVER einen Leerstring.
    # @param retry Anzahl der Wiederholungsversuche nach (10s) wenn Schnittstelle aktiv, aber keine Internetverbindung
    proc inetcheck {{retry 5}} {; #{{{
        variable local_ip
        variable netzwerk_aktiv
        variable internet_verfuegbar
        variable webservice_verfuegbar
        variable after_id

        set ::PORTALSERVER ""
        if {$after_id != ""} {
            after cancel $after_id
        set after_id ""
        }
        if {[localIfcheck nwmsg]} {
            # Lokale Schnittstelle ist aktiv.
            set netzwerk_aktiv 1
            srvLog {inetcheck} Notice "Lokale Netzwerkschnittstelle ist aktiv ($nwmsg)"
            # => Verfuegbarkeit des Internets pruefen
            #    Das geschieht mittels ping auf den Google DNS-Server 8.8.8.8.
            if {![catch {exec ping -c1 8.8.8.8} result]} {
                srvLog {inetcheck} Notice "Internetzugang O.K."
                set internet_verfuegbar 1
                setStationsstatus INET
                # Antwortet der Webservice ?
                if {[execWsRequest ping {} wsAntwort]} {; # Er antwortet
                    srvLog {inetcheck} Debug "Webservice antwortet auf /ping mit \"$wsAntwort\"."
                    # Webservice Version festhalten
                    set dictWsAntwort [::json::json2dict $wsAntwort]
                    if {[dict exists $dictWsAntwort version]} {
                        set ::WS_VERSION [dict get $dictWsAntwort version]
                        srvLog {inetcheck} Notice "Webservice ${::WS_URL} hat die Version ${::WS_VERSION}."
                    }
                    # (Bei der Gelegenheit setzen wir gleich die IPadresse.)
                    if {[execWsRequest set [list station_nr $::STATION_NR what local_ip value $local_ip] wsAntwort]} {
                        srvLog {inetcheck} Debug "Webservice antwortet auf /set local_ip mit \"$wsAntwort\"."
                        if {[regexp {^{.*}$} $wsAntwort]} {; # JSON-Antwort
                            set dictWsAntwort [::json::json2dict $wsAntwort]
                            if {[dict exists $dictWsAntwort portalserver]} {
                                set ::PORTALSERVER [dict get $dictWsAntwort portalserver]
                                srvLog {inetcheck} Notice "Portalserver ist $::PORTALSERVER."
                                setStationsstatus ANGEMELDET
                            } else {
                                srvLog {inetcheck} Warn "Webservice meldet keinen Portalserver."
                                setStationsstatus ANGEMELDET 0
                            }
                        }
                    }
                    set webservice_verfuegbar 1
                    setStationsstatus WEBSERVICE
                } else {; # Webservice antwortet nicht
                    srvLog {inetcheck} Warn "Webservice nicht verfügbar"
                    set webservice_verfuegbar 0
                    setStationsstatus WEBSERVICE 0
                }
            } else {; # Keine Antwort auf ping 8.8.8.8
                srvLog {inetcheck} Warn "Kein Internetzugang"
                set internet_verfuegbar 0
                setStationsstatus INET 0
            }
        } else {; # Fehler
            srvLog {inetcheck} Error $nwmsg
        set netzwerk_aktiv 0
        }
        if {$netzwerk_aktiv && !$internet_verfuegbar && $retry} {
            # Weiter probieren
            set after_id [after 30000 ::wscomm::inetcheck [expr $retry - 1]]
        } else {
            srvLog {inetcheck} Notice "Keine weiteren Verbindungsversuche zum Internet"
        }
        #}}}
    }; # proc inetcheck

    # Verfuegbarkeit des Netzwerkes und des Webservice pruefen
    # Vorsichtshalber etwas warten, falls der Bootvorgang noch nicht so weit ist.
    after 5000 ::wscomm::inetcheck

    #}}}
}; # namespace eval ::wscomm


### FIFO für lokalen Clientzugriff
# Benutzt zum Reagieren auf Änderungen im Netzwerk
# Nur zu experimentellen Zwecken: Loglevel ändern, Deamon abwürgen
#{{{

# Eine über den FIFO eingegangene Meldung baerbeiten
proc handleFifoRequest {} {
    set request [gets $::fd_fifo]
    srvLog FIFO Info "Request eingegangen: $request"
    if {[string equal -length 6 $request "dhcpcd"]} {
        # dhcpcd start/stop
        after 1500 ::wscomm::inetcheck
    } elseif {[string equal -length 5 "$request" "loglevel"]} {
        setLoglevel [regsub {loglevel *} $request {}]
    } elseif {"$request" == "terminate"} {
        srvLog FIFO Notice "\"terminate\" empfangen."
        exit
    }
}; # proc handleFifoRequest 

if {[catch {
        exec /bin/mkdir -p -m 775 /var/local/run/vmkstationd
        if {![file exists /var/local/run/vmkstationd/vmkstationd.fifo]} {
            exec /usr/bin/mkfifo /var/local/run/vmkstationd/vmkstationd.fifo
        }
        set ::fd_fifo [open /var/local/run/vmkstationd/vmkstationd.fifo r+]
        fconfigure $::fd_fifo -blocking 0
        fileevent $::fd_fifo readable handleFifoRequest 
    } result]} {
    srvLog {} Error "FIFO-Kommunikation ist nicht möglich: $result"
} else {
    srvLog {} Notice "Lausche auf /var/local/run/vmkstationd/vmkstationd.fifo"
}

###}}} FIFO für lokalen Clientzugriff


# Globale Daten der Station
#TODO Das kommt nach TTYSattel und TTYHocker
array set ::matten {}


namespace eval ::kernel {; #{{{

    srvLog {::kernel} Notice "::kernel wird initialisiert"

    ### Basis-Kernelmodule für alle Anwendungen aus /lib/kernel laden
    # (Die anwendungsspezifischen Module können sich auf das Vorhandensein dieser Module verlassen.)
    set mods1_loaded [list]; # Die geladenen Module müssen anschließend initialisiert werden.
    foreach tclfile [glob $::dir/lib/kernel/*.tcl] {
        set mod_loaded ""; # Das muß das Modul beim Laden setzen.
        
        if {[catch {source $tclfile} load_error]} {
            srvLog {::kernel} Error "$load_error\n${::errorInfo}"
            continue
        }
        if {"$mod_loaded" == ""} {
            srvLog {::kernel} Error "mod_loaded nicht gesetzt von $tclfile"
        } else {
            lappend mods1_loaded $mod_loaded
            srvLog {::kernel} Notice "Modul $mod_loaded geladen aus $tclfile"
        }
    }
    # Module initialisieren
    foreach mod_loaded $mods1_loaded {
        if {"[info procs ${mod_loaded}::init]" != ""} {
            if {[catch {${mod_loaded}::init} init_error]} {
                srvLog {::kernel} Error $init_error
            } else {
                srvLog {::kernel} Info "Modul $mod_loaded initialisiert"
            }
        } else {
            srvLog {::kernel} Warn "${mod_loaded}::init existiert nicht. Modul wird nicht initialisiert."
        }
    }

    ### Anwendungsspezifische Kernelmodule aus custom/kernel laden
    # (Die Anwendungen können sich auf das Vorhandensein dieser Module verlassen.)
    # (Kernel-)Module laden
    set mods2_loaded [list]; # Die geladenen Module müssen anschließend initialisiert werden.
    #TODO Duplikate vermeiden
    foreach tclfile [glob $::dir/custom/kernel/*.tcl] {
        set mod_loaded ""; # Das muß das Modul beim Laden setzen.
        
        if {[catch {source $tclfile} load_error]} {
            srvLog {::kernel} Error "$load_error\n${::errorInfo}"
            continue
        }
        if {"$mod_loaded" == ""} {
            srvLog {::kernel} Error "mod_loaded nicht gesetzt von $tclfile"
        } else {
            lappend mods2_loaded $mod_loaded
            srvLog {::kernel} Notice "Modul $mod_loaded geladen aus $tclfile"
        }
    }
    # Module initialisieren
    foreach mod_loaded $mods2_loaded {
        if {"[info procs ${mod_loaded}::init]" != ""} {
            if {[catch {${mod_loaded}::init} init_error]} {
                srvLog {::kernel} Error $init_error
            } else {
                srvLog {::kernel} Info "Modul $mod_loaded initialisiert"
            }
        } else {
            srvLog {::kernel} Warn "${mod_loaded}::init existiert nicht. Modul wird nicht initialisiert."
        }
    }

    srvLog {::kernel} Debug "Kernelmodule: [namespace children]"

    # Server und Kernelmodule starten
    proc start {} {; # (neu)
        srvLog {} Notice "Starte Arbeitsstationsdämon auf Port $::CONF_PORT ..."
        if {[catch {
            # HTTP/WS-Server starten
            ::WSServer::start "port $::CONF_PORT doc_root $::dir/htdocs"
            srvLog {} Notice "... gestartet."
            #TODO Alle Basis-Kernelmodule starten, sofern sie eine start-Prozedur haben.
            # Treiberüberwachung starten
            watchTTYACM::start
            srvLog {::kernel} Notice "::kernel gestartet"
        } errmsg]} {
            srvLog {::kernel} Error "::kernel nicht gestartet: $errmsg"
            puts "Kernel konnte nicht gestartet werden. (s. Logfile $::CONF_LOGFILE)"
            exit 1
        }
    }

    #}}}
}; # ::kernel

::kernel::start


### Die Anwendungen
#TODO Doc zu den Konventionen (Variable, Prozeduren)
namespace eval ::apps {; #{{{
    variable apps_loaded [dict create]; # Schlüssel App, Wert Filename

    srvLog "[namespace current]" Notice "::apps werden initialisiert"

    # Fehlermeldung als wsevent in JSON erstellen
    # Die erstellten Schlüssel sind:
    #   "wsevent" mit Wert "error"
    #   "class" Fehlerklasse
    #   "source"  Fehlerquelle
    #   "nr"    Fehlernummer
    #   "msg"   Eine Fehlermeldung
    # Derzeit definierte Fehlernummern 
    #   1 Insufficient data (for analysis)
    #   2 Still recording
    #   3 Unstored recording(s)
    #   4 No session
    #   5 No finished recording
    # @param class_nr   Liste aus Fehlerklasse (internal|client|user) und nr
    # @param msg        Text
    proc createJSONError {class_nr source msg} {; #{{{
        srvLog [namespace current] Debug "createJSONError '$class_nr' '$msg'"
        set class [lindex $class_nr 0]
        set nr [lindex $class_nr 1]
        # Gültige Fehlerklasse?
        if {"$class" ni {internal client user}} {
            srvLog [namespace current] Debug "createJSONError Unknown error class: \"$class\""
        }
        set msg [string map {\" \\\"} $msg]
        set kvlist [list "wsevent" "error" "class" $class "source" "$source" "msg" $msg]
        if {"$nr" != ""} {
            lappend kvlist nr $nr
        }
        set json_error [::kvlist2json $kvlist]
        return $json_error
        #}}}
    }; # proc createJSONError 


    # App neu laden
    # @return Fertigmeldung
    proc reload {app} {; #{{{
        variable apps_loaded
 
        if {[catch {
            # Applikation stoppen, falls gerade am Laufen
            if {[info exists ::apps::${app}::started]} {
                set started [set ::apps::${app}::started]
            } else {
                set started 0
            }
            if {$started} {
                if {"[info procs ::apps::${app}::stop]" != ""} {
                    ::apps::${app}::stop
                    append msg "$app stopped\n"
                }
            }
            set tclfile [dict get $apps_loaded $app]
            set app_loaded ""
            source $tclfile
            if {"$app_loaded" == ""} {
                append msg "Error: app_loaded not set by $tclfile"
            } else {
                append msg "$app_loaded loaded from $tclfile"
                # Initialisieren
                if {"[info procs ${app_loaded}::init]" != ""} {
                    ${app_loaded}::init
                    append msg "\n$app_loaded initialisiert"
                } else {
                    append msg "\n${app_loaded}::init existiert nicht. App wird nicht initialisiert."
                }
                # neu starten, falls vorher auch gestartet
                if {$started} {
                    if {"[info procs ::apps::${app}::start]" != ""} {
                        ::apps::${app}::start
                        append msg "\n$app (re)started"
                    }
                }
            }
        } result]} {
            set msg "Error: $result"
        }
        return $msg
        #}}}
    }; # proc reload 


    ## Apps aus custom/apps laden und initialisieren
    # (Anwendungs-)Module laden
    foreach tclfile [glob $::dir/custom/apps/*.tcl] {
        set app_loaded ""; # Das muß das Modul beim Laden setzen.
        if {[catch {source $tclfile} load_error]} {
            srvLog [namespace current] Error "$load_error\n$errorInfo"
            continue
        }
        if {"$app_loaded" == ""} {
            srvLog {::kernel} Error "app_loaded nicht gesetzt von $tclfile"
        } else {
            dict set apps_loaded $app_loaded $tclfile
            srvLog {::kernel} Notice "$app_loaded geladen aus $tclfile"
        }
    }
    # Apps initialisieren
    foreach app_loaded [dict keys $apps_loaded] {
        if {"[info procs ${app_loaded}::init]" != ""} {
            if {[catch {${app_loaded}::init} init_error]} {
                srvLog [namespace current] Error $init_error
            } else {
                srvLog [namespace current] Info "$app_loaded initialisiert"
            }
        } else {
            srvLog [namespace current] Warn "${app_loaded}::init existiert nicht. App wird nicht initialisiert."
        }
    }

    #}}}
}; # ::apps


# Deamon erfolgreich gestartet
# Falls es einen Link auf eine Fallback-Version gibt, wird der nicht mehr gebraucht.
if {[file exists /usr/local/vmkstationd_last]} {
    file delete /usr/local/vmkstationd_last
    srvLog {} Notice "/usr/local/vmkstationd_last gelöscht."
}

vwait forever
exec stty echo

